
MS_PER_MINUTE = 60 * 1000
MS_PER_DAY = 24 * 60 * MS_PER_MINUTE

# Copy the UTC face of a `Date` object to a new object's local face, or vice versa.

# The special cases ensure that any year is treated as non-rolling; e.g. 99 is
# 0099 and not 1999. (The fact that the shape of a year is congruent modulo 400
# years is used to help minimize the necessary changes.)

# Returns a `Date` object with its UTC face (the face returned by `.getUTC*()`)
# set to the local face (the face returned by the local methods `.get*()`
# corresponding to `.getUTC*()`) suggested by the `Date` `from`.
#
# The conversion is lossless in this direction.
dateLocalToUtc = (from) ->
	y = from.getFullYear()
	mo = from.getMonth()
	d = from.getDate()
	h = from.getHours()
	min = from.getMinutes()
	s = from.getSeconds()
	ms = from.getMilliseconds()
	to = new Date Date.UTC y, mo, d, h, min, s, ms
	if to.getUTCFullYear() isnt y
		to.setTime Date.UTC y + 400, mo, d, h, min, s, ms
		to.setUTCFullYear y
	to

# Returns a `Date` object with its local face set to the UTC face suggested by
# of the `Date` object `from`.
#
# This conversion is generally lossy if the face falls within a daylight/summer
# time ambiguity; the actual results are dependent on the JavaScript
# implementation.
dateUtcToLocalRaw = (from) ->
	y = from.getUTCFullYear()
	mo = from.getUTCMonth()
	d = from.getUTCDate()
	h = from.getUTCHours()
	min = from.getUTCMinutes()
	s = from.getUTCSeconds()
	ms = from.getUTCMilliseconds()
	to = new Date y, mo, d, h, min, s, ms
	if to.getFullYear() isnt y
		to = new Date y + 400, mo, d, h, min, s, ms
		to.setFullYear y
	to

# `faceMs` is an interpretation of an ms date as if the epoch were not
# strictly UTC but entirely timezone-ignorant instead (i.e., the number of ms
# since `1970-01-01T00:00:00` with no specified timezone). If a `Date` object
# is set to `faceMs`, the intended face appears on its UTC face.

# Returns a `Date` object with the local face set as suggested by `faceMs`. 
#
# This conversion is generally lossy if the face falls within a daylight/summer
# time ambiguity; the actual results are dependent on the JavaScript
# implementation.
faceMsToLocalRaw = (faceMs) -> dateUtcToLocalRaw new Date faceMs


# Returns a `Date` object with its local face set to the UTC face suggested by
# of the `Date` object `from`. This conversion is generally lossy if the face
# falls within a daylight/summer time ambiguity; such times are adjusted
# according to the following rules:
#
# - If the face date is skipped by a DST skip forward, the time is resolved to
#   the first minute following the skip (e.g., if 01:59 is directly followed by
#   03:00, 02:30 would be mapped to 03:00).
#	- If two events are scheduled at different times, the event scheduled
#	  earlier will occur no later than the event scheduled later. If one or
#	  both events are scheduled for a skipped time, they may occur at the same
#	  time.
# - If the face date happens twice due a DST skip backward, the time is
#   resolved to the earlier instance (e.g. if 01:00-01:59 daylight time is
#   directly followed by 01:00-01:59 standard time, 01:30 would be mapped to
#   01:30 daylight time).
#	- An event scheduled for a time that is doubled will still occur only once,
#	  at the earlier instance of the time.
#	- A face cannot be resolved to the later instance of a doubled time;
#	  specifically, even if the target application has a concept of *current
#	  time* and the later instance of the time is in the future but the earlier
#	  instance is in the past, the time itself is considered past.

dateUtcToLocalAdjusted = (dateWithUtcFace) ->
	faceMs = dateWithUtcFace.getTime()
	date = dateUtcToLocalRaw dateWithUtcFace
	dateOffset = date.getTimezoneOffset()

	previousDateOffset = (faceMsToLocalRaw faceMs - MS_PER_DAY).getTimezoneOffset()
	
	if faceMs is dateLocalToUtc(date).getTime()
		# Face matches
		if dateOffset is previousDateOffset
			# Matches earliest (or only) instance.
			date
		else
			# If there are two instances, matches the later one and needs adjustment.
			# Otherwise, matches the only instance.
			offsetChangeMinutes = dateOffset - previousDateOffset
			if offsetChangeMinutes <= 0
				# This is a skipping rather than a doubling adjustment; the
				# offset change is the wrong direction
				date
			else
				# The offset change is subtracted from the date's epoch time
				# rather than its face.
				adjustedDate = new Date date.getTime() - (offsetChangeMinutes * MS_PER_MINUTE)

				# If the offsets match, the recent past contains no discontinuity,
				# so there is only one instance. Otherwise, adjust the time to be
				# before the discontinuity.
				if adjustedDate.getTimezoneOffset() is dateOffset then date else adjustedDate
	else
		# Face refers to a skipped time. Resolve to the first minute the end of
		# the discontinuity. The discontinuity is located by binary search;
		# minutes are differentiated based on whether a given time's local
		# offset matches the previous day's offset (before the discontinuity)
		# or not (after the discontinuity).
		followingDateOffset = (faceMsToLocalRaw faceMs + MS_PER_DAY).getTimezoneOffset()
		offsetChangeMinutes = previousDateOffset - followingDateOffset
		faceMinute = faceMs / MS_PER_MINUTE
		before = Math.floor( faceMinute - offsetChangeMinutes )
		after = Math.ceil( faceMinute + offsetChangeMinutes )
		while before + 1 < after
			mid = Math.floor( (before + after) / 2 )
			midOffset = (faceMsToLocalRaw mid * MS_PER_MINUTE).getTimezoneOffset()
			if midOffset is previousDateOffset
				before = mid
			else
				after = mid
		faceMsToLocalRaw after * MS_PER_MINUTE
